#ifndef _ARTUTILS_H
#define _ARTUTILS_H

#include "art.h"
#if defined(sparc)
#include <sys/byteorder.h>
#endif

extern FileOffset errorRoot, errorMsg, remoteStartNode;
extern FileOffset q_position;
extern FILE	*HatFileRandom, *HatFileSeq, *OutputFile, *BridgeFile;
extern char*	progname;			/* browser program name */
extern unsigned	filesize, outputsize;
extern bool	hat_interrupted;

typedef enum {
  Literal, Variable, Constructor, ConstrFields, Abstract
} AtomType;

typedef struct {
  char* srcname;
  int   line;
  int   column;
} SrcRef;

typedef struct {
  char* idname;
  char* modname;
  char* srcname;
  char  fixity;
  char  arity;
  int   defnline;
  int   defncolumn;
  bool  isTraced;
  AtomType  atomtype;
} Ident;

void		initialise	(int argc, char **argv);
void		finalise	(void);
FILE*		openFile	(char* base, char* ext);
int		sizeFile	(char* base, char* ext);
int		freadAt		(FileOffset fo, void* ptr
				,int size, int nmemb, FILE* stream);
FileOffset	readFO		(void);
char*		readString	(void);
int		q_fread		(void* ptr, int size, int nmemb, FILE* stream);
FileOffset	q_readFO	(void);
char*		q_readString	(void);
char		q_peek		(void);
void		readModuleAt	(FileOffset fo, char** modname, char** srcname
				,bool* traced);
Ident*		readAtomAt	(FileOffset fo);
Ident*		readValueAt	(FileOffset fo);
SrcRef*		readSRAt	(FileOffset fo);
void		q_skipNode	(char tag);

FileOffset	readTraceAt	(FileOffset fo, char** expr, SrcRef** sr
				,int* infix,int followHidden,int depth);
char*		infixPrint	( char* str1, int arg1
				, char* strfn, int fn
				, char* str2, int arg2);

#define noFixity      ((char)3)
#define isInfix(fix)  ((fix%4)!=3)
#define isInfixL(fix) ((fix%4)==2)
#define isInfixR(fix) ((fix%4)==1)
#define isInfixN(fix) ((fix%4)==0)
#define priority(fix) (fix/4)

/* prototypes for hat-trail only */
void		openHatFile	(char* prog, char* arg);
void		openBridgeFile	(char* arg);
FileOffset	getBridgeValue	(void);
FileOffset	getErrorLoc	(void);
char*		errorMessage	(void);
char*		versionNumber	(void);
int		getNodeType	(FileOffset fo);
FileOffset	parentNode	(FileOffset fo);
char*		getNm		(FileOffset fo);
char*		getNmMod	(FileOffset fo);
int		getFixity	(FileOffset fo);
bool		isLiteral	(FileOffset fo);
bool		isConstructor	(FileOffset fo);
bool		isConstrFields	(FileOffset fo);
bool		isLambda	(FileOffset fo);
int		getExpArity	(FileOffset fo);
FileOffset	getExpArg	(FileOffset fo, int n);
FileOffset	getFieldLabel	(FileOffset fo, int n);
FileOffset	getSrcRef	(FileOffset fo);
char*		getSrcRefFile	(FileOffset fo);
int		srcRefLine	(FileOffset fo);
int		srcRefCol	(FileOffset fo);
FileOffset	getDefnRef	(FileOffset fo);
char*		getDefnFile	(FileOffset fo);
int		defnLine	(FileOffset fo);
int		defncol		(FileOffset fo);
FileOffset	peekTrace	(FileOffset fo);
FileOffset	getResult	(FileOffset fo, bool stopAtHidden);
FileOffset	getResultNoCycle(FileOffset fo, bool stopAtHidden);
FileOffset	peekResult	(FileOffset fo);


/* prototypes for other tools */
void		ctrlC			(int sig);
int		fileoffset_compare	(FileOffset i, FileOffset j);

#endif
